import 'dart:async';
import 'dart:convert';
import 'dart:typed_data';

import 'package:flutter/foundation.dart';
import 'package:flutterx_http/src/transform_request.dart';
import 'package:flutterx_utils/flutterx_utils.dart';
import 'package:http/http.dart';

part 'x_http_exception.dart';
part 'x_http_extensions.dart';

/// [XHttpClient] interceptor called before the [BaseRequest] is sent.
/// [returns] the request that will be sent. It can be the unmodified/modified input request or either a
/// barely new [BaseRequest]. See [XHttpClient.send]
typedef OnRequest = FutureOr<BaseRequest> Function(BaseRequest request);

/// [XHttpClient] interceptor called after the [StreamedResponse] is received.
/// [returns] the response that will be read into a [Response]. It can be the unmodified/modified input response
/// or either a barely new [StreamedResponse]. See [XHttpClient.send]
typedef OnResponse = FutureOr<StreamedResponse> Function(StreamedResponse response);

/// [XHttpClient] interceptor called after the [Response] is read and the statusCode is not 2xx.
/// [returns] null to throw the [HTTPException] or a new [BaseRequest] that will be sent and it's response will
/// be returned instead of throwing this error. See [XHttpClient.send]
typedef OnError = FutureOr<BaseRequest?> Function(XHTTPException error);

/// Type alias for query parameters map
typedef QueryParameters = Map<String, dynamic>;

/// Type alias for headers map
typedef HeadersMap = Map<String, String>;

/// An HTTP Client wrapper that helps you to send HTTP request
class XHttpClient with Loggable {
  /// The wrapped client factory
  final ValueGetter<Client> clientFactory;

  /// Optional baseUrl for all requests of this client. See [buildUri]
  final String? baseUrl;

  /// Optional [OnRequest] interceptor
  final OnRequest? onRequest;

  /// Optional [OnResponse] interceptor
  final OnResponse? onResponse;

  /// Optional [OnError] interceptor
  final OnError? onError;

  /// Creates a client wrapping [clientFactory], by default [Client] factory is used.
  ///
  /// You can provide a [baseUrl] which will be used for composing urls of this client [applicationJsonRequest]s
  ///
  /// And you can also provide [onRequest], [onResponse], and [onError] interceptors that will be used
  /// when the request is [send]
  XHttpClient(
      {this.clientFactory = _defaultClientFactory, this.baseUrl, this.onRequest, this.onResponse, this.onError});

  /// [send]s an HTTP HEAD [applicationJsonRequest] with the given [headers] to the given [url] with query [params].
  Future<Response> head(String url, {QueryParameters params = const {}, HeadersMap headers = const {}}) =>
      send(applicationJsonRequest('HEAD', buildUri(baseUrl: baseUrl, url: url, params: params), headers: headers));

  /// [send]s an HTTP GET [applicationJsonRequest] with the given [headers] to the given [url] with query [params].
  Future<Response> get(String url, {QueryParameters params = const {}, HeadersMap headers = const {}}) =>
      send(applicationJsonRequest('GET', buildUri(baseUrl: baseUrl, url: url, params: params), headers: headers));

  /// [send]s an HTTP POST [applicationJsonRequest] with the given [headers], [body] and body [encoding] to the given
  /// [url] with query [params].
  Future<Response> post(String url,
          {QueryParameters params = const {}, HeadersMap headers = const {}, Object? body, Encoding? encoding}) =>
      send(applicationJsonRequest('POST', buildUri(baseUrl: baseUrl, url: url, params: params),
          headers: headers, body: body, encoding: encoding));

  /// [send]s an HTTP PUT [applicationJsonRequest] with the given [headers], [body] and body [encoding] to the given
  /// [url] with query [params].
  Future<Response> put(String url,
          {QueryParameters params = const {}, HeadersMap headers = const {}, Object? body, Encoding? encoding}) =>
      send(applicationJsonRequest('PUT', buildUri(baseUrl: baseUrl, url: url, params: params),
          headers: headers, body: body, encoding: encoding));

  /// [send]s an HTTP PATCH [applicationJsonRequest] with the given [headers], [body] and body [encoding] to the given
  /// [url] with query [params].
  Future<Response> patch(String url,
          {QueryParameters params = const {}, HeadersMap headers = const {}, Object? body, Encoding? encoding}) =>
      send(applicationJsonRequest('PATCH', buildUri(baseUrl: baseUrl, url: url, params: params),
          headers: headers, body: body, encoding: encoding));

  /// [send]s an HTTP DELETE [applicationJsonRequest] with the given [headers], [body] and body [encoding] to the given
  /// [url] with query [params].
  Future<Response> delete(String url, {QueryParameters params = const {}, HeadersMap headers = const {}}) =>
      send(applicationJsonRequest('DELETE', buildUri(baseUrl: baseUrl, url: url, params: params), headers: headers));

  /// Sends an HTTP [request] and asynchronously [returns] the response.
  Future<Response> send(BaseRequest request) async {
    Response response;
    if (onRequest != null) request = await onRequest!(request);
    _printRequest(request);
    final client = clientFactory();
    try {
      var stream = await client.send(request);
      if (onResponse != null) stream = await onResponse!(stream);
      response = await Response.fromStream(stream);
    } finally {
      client.close();
    }
    _printResponse(response);
    if (response.statusCode ~/ 100 == 2) return response;
    final error = XHTTPException(response);
    final retry = onError == null ? null : await onError!(error);
    if (retry != null) return send(retry);
    throw error;
  }

  void _printRequest(BaseRequest request) => log('HTTP Request', () {
        logKV('method', request.method);
        logKV('url', '${request.url.scheme}://${request.url.host}:${request.url.port}${request.url.path}');
        final parameters = request.url.queryParameters;
        if (parameters.isNotEmpty) log('parameters:', () => parameters.forEach(logKV));
        if (request.headers.isNotEmpty) log('headers:', () => request.headers.forEach(logKV));
        if (request.contentLength != null) logKV('contentLength', request.contentLength);
        void _printRequestBody(BaseRequest request) {
          if (request is Request) {
            final body = request.body;
            if (body.isNotEmpty) log('body:', () => log(body));
          } else if (request is StreamedRequest) {
            log('body: stream');
          } else if (request is MultipartRequest) {
            log('body: multipart', () {
              if (request.fields.isNotEmpty) log('fields', () => request.fields.forEach(logKV));
              if (request.files.isNotEmpty) log('files', () => request.files.map((e) => e.filename).forEach(log));
            });
          } else if (request is TransformRequest) {
            logKV('transform', true);
            _printRequestBody(request.source);
          }
        }

        _printRequestBody(request);
      });

  void _printResponse(BaseResponse response) => log('HTTP Response', () {
        final request = response.request;
        if (request != null) {
          logKV('method', request.method);
          logKV('url', request.url);
        }
        logKV('statusCode', response.statusCode);
        logKV('reasonPhrase', response.reasonPhrase);
        if (response.headers.isNotEmpty) log('headers:', () => response.headers.forEach(logKV));
        if (response is Response) {
          final body = response.body;
          if (body.isNotEmpty) log('body:', () => log(body));
        } else if (response is StreamedResponse) {
          log('body: stream');
        }
      });

  /// Creates an HTTP application/json request
  ///
  /// The content-type of the request will default to 'application/json' if not set.
  ///
  /// If provided, [body] sets the body of the request. It can be a [String], any [Object].
  ///
  /// If [body] is a String, it's encoded using [encoding] and used as the body of the request
  /// else it's encoded to a json String through [json.encode] and treated as a String
  ///
  /// [encoding] defaults to [utf8].
  static Request applicationJsonRequest(String method, Uri uri,
      {HeadersMap headers = const {}, Object? body, Encoding? encoding}) {
    final request = Request(method, uri);
    if (headers.isNotEmpty) request.headers.addAll(headers);
    request.headers.putIfAbsent('content-type', () => 'application/json');
    if (encoding != null) request.encoding = encoding;
    if (body != null) request.body = body is String ? body : json.encode(body);
    return request;
  }

  /// Build an uri from [baseUrl], [url] and query [params]
  ///
  /// Note: if [baseUrl] is not set or [url] does not start with '/', only [url] will be used.
  ///
  /// E.g. buildUri(baseUrl: 'https://my.domain.com', url: '/api/myEndPoint', params: {'q': 'hello'})
  ///      -> 'https://my.domain.com/api/myEndPoint?q=hello'
  static Uri buildUri({String? baseUrl, required String url, QueryParameters params = const {}}) {
    if (baseUrl != null && url.startsWith('/'))
      url = baseUrl.endsWith('/') ? baseUrl + url.substring(1) : baseUrl + url;
    final uri = Uri.parse(url);
    return params.isEmpty ? uri : uri.appendQueryParameters(params);
  }

  static Client _defaultClientFactory() => Client();
}
