part of 'x_http_client.dart';

extension UriExt on Uri {
  /// Append [QueryParameters] to this [Uri]
  Uri appendQueryParameters(QueryParameters params) {
    final queryParameters = Map<String, String>.from(this.queryParameters);
    queryParameters.addAll(params.map((key, value) => MapEntry(key, value.toString())));
    return Uri(
        scheme: scheme,
        userInfo: userInfo,
        host: host,
        port: port,
        path: path,
        queryParameters: queryParameters,
        fragment: hasFragment ? fragment : null);
  }
}

extension BaseRequestExt on BaseRequest {
  /// Copy a [BaseRequest] and optionally change parameters
  BaseRequest copy({
    String? method,
    Uri? url,
    bool? persistentConnection,
    bool? followRedirects,
    int? maxRedirects,
    Map<String, String>? headers,
  }) {
    if (this is Request)
      return (this as Request).copy(
          method: method,
          url: url,
          persistentConnection: persistentConnection,
          followRedirects: followRedirects,
          maxRedirects: maxRedirects,
          headers: headers);
    if (this is StreamedRequest)
      return (this as StreamedRequest).copy(
          method: method,
          url: url,
          persistentConnection: persistentConnection,
          followRedirects: followRedirects,
          maxRedirects: maxRedirects,
          headers: headers);
    if (this is MultipartRequest)
      return (this as MultipartRequest).copy(
          method: method,
          url: url,
          persistentConnection: persistentConnection,
          followRedirects: followRedirects,
          maxRedirects: maxRedirects,
          headers: headers);
    throw UnsupportedError('unsupported request of type $runtimeType');
  }
}

extension RequestExt on Request {
  /// Copy a [Request] and optionally change parameters
  Request copy({
    String? method,
    Uri? url,
    bool? persistentConnection,
    bool? followRedirects,
    int? maxRedirects,
    Map<String, String>? headers,
    String? body,
    List<int>? bodyBytes,
    Map<String, String>? bodyFields,
    Encoding? encoding,
  }) {
    final result = Request(method ?? this.method, url ?? this.url)
      ..persistentConnection = persistentConnection ?? this.persistentConnection
      ..followRedirects = followRedirects ?? this.followRedirects
      ..maxRedirects = maxRedirects ?? this.maxRedirects
      ..headers.addAll(headers ?? this.headers)
      ..encoding = encoding ?? this.encoding;
    if (body != null)
      result.body = body;
    else if (bodyBytes != null)
      result.bodyBytes = bodyBytes;
    else if (bodyFields != null)
      result.bodyFields = bodyFields;
    else
      result.bodyBytes = this.bodyBytes;
    return result;
  }
}

extension StreamedRequestExt on StreamedRequest {
  /// Copy a [StreamedRequest] and optionally change parameters
  StreamedRequest copy({
    String? method,
    Uri? url,
    int? contentLength = -1,
    bool? persistentConnection,
    bool? followRedirects,
    int? maxRedirects,
    Map<String, String>? headers,
    Stream<List<int>>? body,
  }) {
    final result = StreamedRequest(method ?? this.method, url ?? this.url)
      ..contentLength = contentLength != -1 ? contentLength : this.contentLength
      ..persistentConnection = persistentConnection ?? this.persistentConnection
      ..followRedirects = followRedirects ?? this.followRedirects
      ..maxRedirects = maxRedirects ?? this.maxRedirects
      ..headers.addAll(headers ?? this.headers);
    (body ?? finalize())
        .listen(result.sink.add, onError: result.sink.addError, onDone: result.sink.close, cancelOnError: true);
    return result;
  }
}

extension MultipartRequestExt on MultipartRequest {
  /// Copy a [MultipartRequest] and optionally change parameters
  MultipartRequest copy({
    String? method,
    Uri? url,
    bool? persistentConnection,
    bool? followRedirects,
    int? maxRedirects,
    Map<String, String>? headers,
    Map<String, String>? fields,
    List<MultipartFile>? files,
  }) {
    final result = MultipartRequest(method ?? this.method, url ?? this.url)
      ..persistentConnection = persistentConnection ?? this.persistentConnection
      ..followRedirects = followRedirects ?? this.followRedirects
      ..maxRedirects = maxRedirects ?? this.maxRedirects
      ..headers.addAll(headers ?? this.headers)
      ..fields.addAll(fields ?? this.fields)
      ..files.addAll(files ?? this.files);
    return result;
  }
}
